import factory
import faker

from regex_match.objects.parsers import UrlParser

faker = faker.Factory.create()


class UrlParserMethodFactory(factory.Factory):
    class Meta:
        model = UrlParser

    url = faker.url()


class ModelRuleFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'regex_match.ModelRule'


class ModelExceptionFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'regex_match.ModelException'

    name = faker.name()
    template = '{scheme}://{host}{path}?v={query}'
    model_field = 'url'
    model_field_type = 'url'


class ExceptionParserFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'regex_match.ExceptionParser'
    model_exception = factory.SubFactory('regex_match.ModelException')


class MatchingRuleFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'regex_match.MatchingRule'

    model_rule = factory.SubFactory(ModelRuleFactory)


class ParserMethodFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'regex_match.ParserMethod'
        django_get_or_create = ('name',)


class RegexRuleFactory(factory.DjangoModelFactory):
    class Meta:
        model = 'regex_match.RegexRule'

    matching_rule = factory.SubFactory(MatchingRuleFactory)
    parser_method = factory.SubFactory(ParserMethodFactory)



from django.core.cache import cache
from django.db.models import Q
from funcy import first
from toolz import curry
from toolz.curried import filter, pipe

from regex_match.models.model_rule import ModelRule


class ModelMatcher(object):

    def __init__(self, model, model_field, obj):
        self.obj = obj
        self.model = model
        self.model_field = model_field

    def get(self):
        """
        1) Returns the first object that have a matching matching_rule
        2) A matching_rule is matching if all regex rules are matching
        """

        all_regex_match = self._build_match(
            predicate=self._regex_match,
            fn=all
        )

        regex_predicate = self._match_rule(
            parser_method='regex_rules',
            match=all_regex_match
        )

        any_classification_match = self._build_match(
            predicate=regex_predicate,
            fn=any
        )

        classification_predicate = self._match_rule(
            parser_method='matching_rules',
            match=any_classification_match
        )

        if not isinstance(self.model, str):
            cache_key = hash(self.model) + hash(self.model_field)
            model_rules = cache.get(cache_key)

            if model_rules is None:
                model_rules = self._get_model_rules()
                cache.set(cache_key, model_rules, 30)
        else:
            model_rules = self._get_model_rules()

        return pipe(model_rules, filter(classification_predicate), first)

    @staticmethod
    @curry
    def _match_rule(model_rule, match, parser_method):
        cache_key = hash(model_rule) + hash(match) + hash(parser_method)
        sequence = cache.get(cache_key)

        if sequence is None:
            sequence = list(getattr(model_rule, parser_method).select_related())
            cache.set(cache_key, sequence, 30)

        return match(
            sequence=sequence
        )

    @staticmethod
    @curry
    def _build_match(fn, predicate, sequence):
        """
        Generate match using a fn, a predicate and a sequence of rules
        Returns True, False or an Object depending of the fn
        """
        return fn(
            predicate(item)
            for item in sequence
        )

    def _regex_match(self, rule):
        return rule.is_matching(self.obj)

    def _get_model_rules(self):
        if (type(self.model) == str):
            model = self.model.lower()
        else:
            model = self.model.__name__.lower()

        return list(ModelRule.objects.filter(
            Q(polymorphic_ctype__model=model) &
            Q(model_field=self.model_field)
        ).select_related().prefetch_related())



from django.db import models
from polymorphic.models import PolymorphicModel


class ModelRule(PolymorphicModel):
    model_field = models.CharField(max_length=16)
    model_field_type = models.CharField(max_length=16)
    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)
    priority = models.PositiveIntegerField(default=0)

    class Meta:
        ordering = ('-priority',)

    def is_valid(self):
        return any(
            self.has_regex_rule(item)
            for item in self.matching_rules.all()
        )

    @staticmethod
    def has_regex_rule(matching_rule):
        return matching_rule.is_valid()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('regex_match', '0005_auto_20151117_2133'),
    ]

    operations = [
        migrations.AlterField(
            model_name='modelexception',
            name='template',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='parsermethod',
            name='name',
            field=models.CharField(max_length=100),
        ),
        migrations.AlterField(
            model_name='regexrule',
            name='regex',
            field=models.TextField(),
        ),
    ]



from django.db import models

from .model_rule import ModelRule


class ModelException(ModelRule):
    name = models.CharField(max_length=32)
    template = models.TextField(null=False)

    def format_template(self, obj):
        parsers_dic = self._parsers_to_dic(obj)

        return self.template.format(**parsers_dic)

    def _parsers_to_dic(self, obj):
        parsers = self.parsers.all()

        return dict([self._compile_parser(parser, obj) for parser in parsers])

    def _compile_parser(self, parser, obj):
        return parser.name, parser.parse_obj(obj)



from django.db import models


class MatchingRule(models.Model):
    model_rule = models.ForeignKey(
        "ModelRule", on_delete=models.CASCADE, related_name="matching_rules"
    )
    description = models.CharField(max_length=32)
    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)

    def __str__(self):
        return "{} match [{}]".format(self.model_rule, self._get_regex_rules())

    def _get_regex_rules(self):
        matches = self.regex_rules.all()
        matches_map = map(self._get_match_rule_name, matches)

        return ", ".join(list(matches_map))

    def _get_match_rule_name(self, match):
        return "{} | {}".format(match.parser_method.name, match.regex)

    def is_valid(self):
        return len(self.regex_rules.all()) > 0



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('regex_match', '0002_modelrule_priority'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='modelrule',
            options={'ordering': ('-priority',)},
        ),
    ]



import re

from django.db import models

from .parser import Parser


class RegexRule(Parser):
    matching_rule = models.ForeignKey(
        "MatchingRule", on_delete=models.CASCADE, related_name="regex_rules"
    )
    regex = models.TextField(null=False)

    def __str__(self):
        return "{} match {}".format(self.method(), self.regex)

    def is_matching(self, obj):
        """
        Test if a regex match a rule's method
        Returns True of False
        """
        matching = False
        regex = re.compile(self.regex)
        string = self.parse_obj(obj)
        if string is not None:
            matching = bool(regex.match(string))
        return matching
